from list.linkedListBasic import *

class LinkedStack:
	def __init__(self):
		self.__list = LinkedListBasic()

	def push(self, newItem):
		self.__list.insert(0, newItem)

	def pop(self):
		return self.__list.pop(0)

	def top(self):
		if self.isEmpty():
			return None
		else:
			return self.__list.get(0)

	def isEmpty(self) -> bool:
		return self.__list.isEmpty()

	def popAll(self):
		self.__list.clear()

	def printStack(self):
		print("Stack from top:", end = '')
		for i in range(self.__list.size()):
			print(self.__list.get(i), end = '')
		print()

# 코드 6-13



from list.linkedListBasic import *

list = LinkedListBasic()
list.append(30); list.insert(0, 20)
a = LinkedListBasic()
a.append(4); a.append(3); a.append(3); a.append(2); a.append(1)
list.extend(a)
list.reverse()
list.pop(0)
print("count(3):", list.count(3))
print("get(2):", list.get(2))
list.printList()

# 코드 5-16


from sort.countingSort import *

def main():
    A = [35, 24, 16, 21, 4, 72, 23, 9, 23, 14, 58, 12, 0]
    print("A[]:       ", A)
    print("Sorted A[]:", countingSort(A))

if __name__ == "__main__":
    main()



from list.circularDoublyLinkedList import *

list = CircularDoublyLinkedList()
list.append(30); list.insert(0, 20)
a = [4, 3, 3, 2, 1]
list.extend(a)
list.reverse()
list.pop(0)
list.sort()
print("count(3):", list.count(3))
print("get(2):", list.get(2))
list.printList()

# 코드 5-26


class AVLNode:
	def __init__(self, newItem, left, right, h):
		self.item = newItem
		self.left = left
		self.right = right
		self.height = h

class AVLTree:
	def __init__(self):
		self.NIL = AVLNode(None, None, None, 0)
		self.__root = self.NIL
		self.LL = 1; self.LR = 2; self.RR = 3; self.RL = 4
		self.NO_NEED = 0; 
		self.ILLEGAL = -1

	# [알고리즘 10-1] 구현: 검색
	def search(self, x):
		return self.__searchItem(self.__root, x)
	
	def __searchItem(self, tNode:AVLNode, x) -> AVLNode:
		if tNode == self.NIL:
			return self.NIL
		elif x == tNode.item:
			return tNode
		elif x < tNode.item:
			return self.__searchItem(tNode.left, x)
		else:
			return self.__searchItem(tNode.right, x)
		
	# [알고리즘 10-3] 구현: 삽입
	def insert(self, x):
		self.__root = self.__insertItem(self.__root, x)
  
	def __insertItem(self, tNode:AVLNode, x) -> AVLNode:
		if tNode == self.NIL: # insert after a leaf(or into an empty tree)
			tNode = AVLNode(x, self.NIL, self.NIL, 1)
		elif x < tNode.item:  # branch left
			tNode.left = self.__insertItem(tNode.left, x)
			tNode.height = 1 + max(tNode.right.height, tNode.left.height)
			type = self.__needBalance(tNode)
			if (type != self.NO_NEED):
				tNode = self.__balanceAVL(tNode, type)
		else:  # branch right
			tNode.right = self.__insertItem(tNode.right, x)
			tNode.height = 1 + max(tNode.right.height, tNode.left.height)
			type = self.__needBalance(tNode)
			if type != self.NO_NEED:
				tNode = self.__balanceAVL(tNode, type)
		return tNode

	# [알고리즘 10-7] 구현: 삭제
	def delete(self, x):
		self.__root = self.__deleteItem(self.__root, x)
	def __deleteItem(self, tNode:AVLNode, x) -> AVLNode:
		if tNode == self.NIL:
			return self.NIL	# item not found!
		else:
			if x == tNode.item: # item found at tNode
				tNode = self.__deleteNode(tNode)
			elif x < tNode.item:
				tNode.left = self.__deleteItem(tNode.left, x)
				tNode.height = 1 + max(tNode.right.height, tNode.left.height)
				type = self.__needBalance(tNode)
				if type != self.NO_NEED:
					tNode = self.__balanceAVL(tNode, type)
			else:
				tNode.right = self.__deleteItem(tNode.right, x)
				tNode.height = 1 + max(tNode.right.height, tNode.left.height)
				type = self.__needBalance(tNode)
				if type != self.NO_NEED:
					tNode = self.__balanceAVL(tNode, type)
			return tNode
  
	def __deleteNode(self, tNode:AVLNode) -> AVLNode:
  		# 3가지 case
  		#     1. tNode이 리프 노드
  		#     2. tNode이 자식이 하나만 있음
  		#     3. tNode이 자식이 둘 있음
		if tNode.left == self.NIL and tNode.right == self.NIL: # case 1(자식이 없음)
			return self.NIL
		elif tNode.left == self.NIL:  # case 2(오른자식뿐)
			return tNode.right
		elif tNode.right == self.NIL: # case 2(왼자식뿐)
			return tNode.left
		else: # case 3(두 자식이 다 있음)
			(rtnItem, rtnNode) = self.__deleteMinItem(tNode.right)
			tNode.item = rtnItem
			tNode.right = rtnNode
			tNode.height = 1 + max(tNode.right.height, tNode.left.height)
			type = self.__needBalance(tNode)	
			if type != self.NO_NEED:
				tNode = self.__balanceAVL(tNode, type)
			return tNode  # tNode survived
  
	def __deleteMinItem(self, tNode:AVLNode) -> tuple:
		if tNode.left == self.NIL:
			# found min at tNode
			return (tNode.item, tNode.right)
		else: # keep branching left, then backtrack
			(rtnItem, rtnNode) = self.__deleteMinItem(tNode.left)
			tNode.left = rtnNode
			tNode.height = 1 + max(tNode.right.height, tNode.left.height) 
			type = self.__needBalance(tNode)
			if type != self.NO_NEED:
				tNode = self.__balanceAVL(tNode, type)
			return (tNode, rtnItem)
 	
	# 균형 잡기 
	def __balanceAVL(self, tNode:AVLNode, type:int) -> AVLNode:
		returnNode = self.NIL
		if type == self.LL:
			returnNode = self.__rightRotate(tNode)
		elif type == self.LR:
			tNode.left = self.__leftRotate(tNode.left)
			returnNode = self.__rightRotate(tNode)
		elif type == self.RR:
			returnNode = self.__leftRotate(tNode)
		elif type == self.RL:
			tNode.right = self.__rightRotate(tNode.right)
			returnNode = self.__leftRotate(tNode)
		else:
			print("Impossible type! Should be one of LL, LR, RR, RL")
		return returnNode
  
	# [알고리즘 11-1] 구현 : 좌회전
	def __leftRotate(self, t:AVLNode) -> AVLNode:
		RChild = t.right
		if RChild == self.NIL:
			print(t.item, "'s RChild shouldn't be NIL!")
		RLChild = RChild.left
		RChild.left = t
		t.right = RLChild
		t.height = 1 + max(t.left.height, t.right.height)
		RChild.height = 1 + max(RChild.left.height, RChild.right.height)
		return RChild

	# [알고리즘 11-1] 구현 : 우회전
	def __rightRotate(self, t:AVLNode) -> AVLNode:
		LChild = t.left
		if LChild == self.NIL:
			print(t.item, "'s LChild shouldn't be NIL!")
		LRChild = LChild.right
		LChild.right = t
		t.left = LRChild
		t.height = 1 + max(t.left.height, t.right.height)
		LChild.height = 1 + max(LChild.left.height, LChild.right.height)
		return LChild

	def __needBalance(self, t:AVLNode) -> int:
		type = self.ILLEGAL
		if (t.left.height + 2 <= t.right.height): # R 유형
			if (t.right.left.height) <= t.right.right.height:
				type = self.RR
			else:
				type = self.RL
		elif (t.left.height) >= t.right.height + 2:  # L 유형
			if (t.left.left.height) >= t.left.right.height:
				type = self.LL
			else:
				type = self.LR
		else:
			type = self.NO_NEED
		return type

	# 기타 
	def isEmpty(self) -> bool:
		return self.__root == self.NIL
 
	def clear(self):
		self.__root = self.NIL

# 코드 11-1


from sort.heapSort import *

def main():
    print("Heapsort!")
    A = [3, 8, 2, 4, 8, 1, 2, 0, 5, 9]
    print("A[]:       ", A)
    heapSort(A)
    print("Sorted A[]:", A)

if __name__ == "__main__":
    main()



from sort.mergeSort import *

def main():
    print("Mergesort test")
    A = [3, 8, 2, 24, 8, 1, 12,1, 0, 55, 0, 256, 333, 23, 65]
    print("A[]:       ", A)
    mergeSort(A, 0, len(A)-1)
    print("Sorted A[]:", A)

if __name__ == "__main__":
    main()


def mergeSort(A, p:int, r:int):
	if p < r:
		q = (p+r) // 2
		mergeSort(A, p, q)
		mergeSort(A, q+1, r)
		merge(A, p, q, r)

def merge(A, p:int, q:int, r:int):
	i = p; j = q+1; t = 0
	tmp = [0 for i in range(len(A))]
	while i <= q and j <= r:
		if A[i] <= A[j]:
			tmp[t] = A[i]; t += 1; i += 1
		else:
			tmp[t] = A[j]; t += 1; j += 1
	while i <= q:
		tmp[t] = A[i]; t += 1; i += 1
	while j <= r:
		tmp[t] = A[j]; t += 1; j += 1
	i = p; t = 0
	while i <= r:
		A[i] = tmp[t]; t += 1; i += 1

# 코드 9-4


def heapSort(A):
	buildHeap(A)
	for last in range(len(A)-1, 0, -1):
		A[last], A[0] = A[0], A[last]
		percolateDown(A, 0, last-1)

def buildHeap(A): # 리스트 A[0...len(A)-1]을 힙으로 만든다
	for i in range((len(A)-2) // 2, -1, -1):
		percolateDown(A, i, len(A)-1)

# A[k]를 루트로 하는 서브 트리가 A[k...end] 범위 내에서 힙 특성을 만족하도록 수선
# 주어진 조건: A[k]의 두 자식을 루트로 하는 서브 트리는 힙 특성을 만족함
def percolateDown(A, k:int, end:int):
	child = 2*k+1	 # 왼자식
	right = 2*k+2	 # 오른자식
	if child <= end:
			if right <= end and A[child] < A[right]:
					child = right
			# child: A[2k+1]와 A[2k+2] 중에 큰 원소의 인덱스
			if A[k] < A[child]:
				A[k], A[child] = A[child], A[k]
				percolateDown(A, child, end)

# 코드 9-7


import math
from sort.insertionSort import insertionSort

def bucketSort(A):  # [0, 1) 범위의 실수 정렬
	n = len(A)
	B = [[] for _ in range(n)]
	for i in range(n):
		B[math.floor(n*A[i])].append(A[i])
	A.clear()
	for i in range(n):
		insertionSort(B[i])
		A.extend(B[i])

# 코드 9-11


from list.circularLinkedList import *

list = CircularLinkedList()
list.append(30); list.insert(0, 20)
a = [4, 3, 3, 2, 1]
list.extend(a)
list.reverse()
re = list.remove(30)
print(re)
list.pop(0)
print("count(3):", list.count(3))
print("get(2):", list.get(2))
list.printList()

# 코드 5-23


from list.listNode import ListNode
from typing import Tuple

class LinkedListBasic:
	def __init__(self):
		self.__head = ListNode('dummy', None)
		self.__numItems = 0

	# [알고리즘 5 - 2] 구현: 연결 리스트에 원소 삽입하기(더미 헤드를 두는 대표 버전)
	def insert(self, i:int, newItem):
		if i >= 0 and i <= self.__numItems:
			prev = self.__getNode(i - 1)
			newNode = ListNode(newItem, prev.next)
			prev.next = newNode
			self.__numItems += 1
		else:
			print("index", i, ": out of bound in insert()") # 필요 시 에러 처리
 
	def append(self, newItem):
		prev = self.__getNode(self.__numItems - 1)
		newNode = ListNode(newItem, prev.next)
		prev.next = newNode
		self.__numItems += 1

	# [알고리즘 5-3] 구현: 연결 리스트의 원소 삭제하기
	def pop(self, i:int):   # i번 노드 삭제. 고정 파라미터
		if (i >= 0 and i <= self.__numItems-1):
			prev = self.__getNode(i - 1)
			curr = prev.next
			prev.next = curr.next
			retItem = curr.item
			self.__numItems -= 1
			return retItem
		else:
			return None
	
	# [알고리즘 5 -4] 구현: 연결 리스트의 원소 x 삭제하기 (더미 헤드를 두는 대표 버전)
	def remove(self, x):
		(prev, curr) = self.__findNode(x)
		if curr != None:
			prev.next = curr.next
			self.__numItems -= 1
			return x
		else:
			return None

	# [알고리즘 5 - 5] 구현: 연결 리스트의 i번 원소 알려주기
	def get(self, i:int):
		if self.isEmpty():
			return None
		if (i >= 0 and i <= self.__numItems - 1):
			return self.__getNode(i).item
		else:
			return None
 
	# [알고리즘 5 -7] 구현: x가 연결 리스트의 몇 번 원소인지 알려주기
	def index(self, x) -> int:
		curr = self.__head.next	 # 0번 노드:  더미 헤드 다음 노드
		for index in range(self.__numItems):
			if curr.item == x:
				return index
			else:
				curr = curr.next
		return -2 # 안 쓰는 인덱스

	# [알고리즘 5 -8] 구현: 기타 작업들
	def isEmpty(self) -> bool:
		return self.__numItems == 0

	def size(self) -> int:
		return self.__numItems

	def clear(self):
		self.__head = ListNode("dummy", None)
		self.__numItems = 0

	def count(self, x) -> int:
		cnt = 0
		curr = self.__head.next  # 0번 노드
		while curr != None:
			if curr.item == x:
					cnt += 1
			curr = curr.next
		return cnt

	def extend(self, a): # 여기서 a는 self와 같은 타입의 리스트
		for index in range(a.size()):
			self.append(a.get(index))
 
	def copy(self):
		a = LinkedListBasic()
		for index in range(self.__numItems):
			a.append(self.get(index))
		return a

	def reverse(self):
		a = LinkedListBasic()
		for index in range(self.__numItems):
			a.insert(0, self.get(index))
		self.clear()
		for index in range(a.size()):
			self.append(a.get(index))

	def sort(self) -> None:
		a = []
		for index in range(self.__numItems):
			a.append(self.get(index))
		a.sort()
		self.clear()
		for index in range(len(a)):
			self.append(a[index])
 
	def __findNode(self, x) -> (Tuple[ListNode, ListNode]):
		prev = self.__head  # 더미 헤드
		curr = prev.next    # 0번 노드
		while curr != None:
			if curr.item == x:
				return (prev, curr)
			else:
				prev = curr; curr = curr.next
		return (None, None)

	# [알고리즘 5-6] 구현: 연결 리스트의 i번 노드 알려주기
	def __getNode(self, i:int) -> ListNode:
		curr = self.__head # 더미 헤드, index: -1
		for index in range(i+1):
			curr = curr.next
		return curr

	def printList(self):
		curr = self.__head.next # 0번 노드: 더미 헤드 다음 노드
		while curr != None:
			print(curr.item, end = ' ')
			curr = curr.next
		print()
	
# 코드 5-15


from heap_.heap import *

# print('Heap!')
h1 = Heap([1, 11, 9, 2, 3])
h1.buildHeap()
h1.clear()
# h1.heapPrint()
h1.insert(7)
h1.insert(5)
h1.insert(9)
h1.insert(4)
h1.insert(11)
h1.insert(19)
h1.insert(20)
h1.insert(21)
h1.insert(11)
# h1.heapPrint()
h1.deleteMax()
# h1.heapPrint()

# 코드 8-9


class ListStack:
	def __init__(self):
		self.__stack = []

	def push(self, x):
		self.__stack.append(x)

	def pop(self):
		return self.__stack.pop()

	def top(self):
		if self.isEmpty():
			return None
		else:
			return self.__stack[-1]

	def isEmpty(self) -> bool:
		return not bool(self.__stack)

	def popAll(self):
		self.__stack.clear() 

	def printStack(self):
		print("Stack from top:", end = ' ')
		for i in range(len(self.__stack)-1, -1, -1):
			print(self.__stack[i], end = ' ')
		print()

# 코드 6-6


class Heap:
	def __init__(self, *args):
		if len(args) != 0:
			self.__A = args[0] # 파라미터로 온 리스트
		else:
			self.__A = []
 
	# [알고리즘 8-2] 구현: 힙에 원소 삽입하기(재귀 알고리즘 버전)
	def insert(self, x):
		self.__A.append(x)
		self.__percolateUp(len(self.__A)-1)

	# 스며오르기
	def __percolateUp(self, i:int):
		parent = (i - 1) // 2
		if i > 0 and self.__A[i] > self.__A[parent]:
			self.__A[i], self.__A[parent] = self.__A[parent], self.__A[i]
			self.__percolateUp(parent)

	# [알고리즘 8-2] 구현: 힙에서 원소 삭제하기
	def deleteMax(self):
		# heap is in self.__A[0...len(self.__A)-1]
		if (not self.isEmpty()):
			max = self.__A[0]
			self.__A[0] = self.__A.pop() # *.pop(): 리스트의 끝원소 삭제 후 원소 리턴
			self.__percolateDown(0)
			return max
		else:
			return None

	# 스며내리기
	def __percolateDown(self, i:int):
		# Percolate down w/ self.__A[i] as the root
		child = 2 * i + 1  # left child
		right = 2 * i + 2  # right child
		if (child <= len(self.__A)-1):
			if (right <= len(self.__A)-1 and self.__A[child] < self.__A[right]):
				child = right  # index of larger child
			if self.__A[i] < self.__A[child]:
				self.__A[i], self.__A[child] = self.__A[child], self.__A[i]
				self.__percolateDown(child)

	def max(self):
		return self.__A[0]

	# 힙 만들기
	def buildHeap(self):
		for i in range((len(self.__A) - 2) // 2, -1, -1):
			self.__percolateDown(i)

	# 힙이 비었는지 확인하기
	def isEmpty(self) -> bool:
		return len(self.__A) == 0

	def clear(self):
		self.__A = []

	def size(self) -> int:
		return len(self.__A)

# 코드 8-8


def quickSort(A, p:int, r:int):
	if p < r:
		q = partition(A, p, r)	 # 분할
		quickSort(A, p, q-1)	 # 왼쪽 부분 리스트 정렬
		quickSort(A, q+1, r)	 # 오른쪽 부분 리스트 정렬

def partition(A, p:int, r:int) -> int:
	x = A[r]					# x: 기준 원소
	i = p-1					# i: 1구역의 끝 지점
	for j in range(p, r):	# j: 3구역의 시작 지점
		if A[j] < x:
			i += 1
			A[i], A[j] = A[j], A[i]  # 교환
	A[i+1], A[r] = A[r], A[i+1]	   # 기준 원소와 2구역 첫 원소 교환
	return i+1

# 코드 9-5


import math

def radixSort(A):
	maxValue = max(A)
	numDigits = math.ceil(math.log10(maxValue))  # 자릿수 계산
	bucket = [[] for _ in range(10)]  # 0, 1, ..., 9에 대한 10개의 리스트
	for i in range(numDigits):
		for x in A:
			y = (x // 10**i) % 10
			bucket[y].append(x)
		A.clear()
		for j in range(10):
			A.extend(bucket[j])
			bucket[j].clear()

# 코드 9-10


from BST.AVLTree import AVLTree
bst1 = AVLTree()
bst1.insert(10)
bst1.insert(20)
bst1.insert(5)
bst1.insert(80)
bst1.insert(90)
bst1.insert(7550)
bst1.insert(30)
bst1.insert(77)
bst1.insert(15)
bst1.insert(40)
bst1.delete(7550)
bst1.delete(10)


# 코드 11-2


class ListQueue:
	def __init__(self):
		self.__queue = []

	def enqueue(self, x):
		self.__queue.append(x)

	def dequeue(self):
		return self.__queue.pop(0) # .pop(0): 리스트의 첫 원소를 삭제한 후 원소 리턴

	def front(self):
		if self.isEmpty():
			return None
		else:
			return self.__queue[0]

	def isEmpty(self) -> bool:
		return (len(self.__queue) == 0);
 
	def dequeueAll(self):
		self.__queue.clear()

	def printQueue(self):
		print("Queue from front:", end = ' ')
		for i in range(len(self.__queue)):
			print(self.__queue[i], end = ' ')
		print()

# 코드 7-6


from hashTable.listHashTable import *

def main():
    print("List Hash Table Demo!")
    h = ListHashTable(11)
    h.insert(10)
    h.insert(21)
    #h.delete(20)   # 아무 영향 없어야 함
    h.insert(20)
    h.insert(5)
    h.insert(80)
    h.insert(32)
    h.printAll()

    h.delete(20)
    h.delete(80)

    # Just for checking purpose
    item = 21
    (i, j) = h.search(item)
    if i == None:
        print("Search Failed!")
    else:
        print("Search:", item, "List:", i, ", item index", j)

if __name__ == "__main__":
    main()


def shellSort(A):  # A[0...n-1]: 정렬할 리스트
	H = gapSequence(len(A))
	for h in H:  # H = [h0, h1, ..., 1]: 갭 수열
		for k in range(h):
			stepInsertionSort(A, k, h)

def stepInsertionSort(A, k:int, h:int):  # A[k, k+h, k+2h, ...]을 정렬한다
	for i in range(k + h, len(A), h):
		j = i - h
		newItem = A[i]
		# 이 지점에서 A[..., j-2h, j-h, j]는 이미 정렬되어 있는 상태임
		# A[..., j-2h, j-h, j, j+h]의 맞는 곳에 A[j+h]를 삽입한다
		while 0 <= j and newItem < A[j]:
			A[j + h] = A[j]
			j -= h
		A[j + h] = newItem

def gapSequence(n:int) -> list: # 갭 수열 만들기. 다양한 선택이 있음
	H = [1]; gap = 1
	while gap < n/5:
		gap = 3 * gap + 1
		H.append(gap)
	H.reverse()
	return H

# 코드 9-8


# 책에는 포함되지 않음
class ListHashTable:
    def __init__(self, n:int):
        self.__table = [[] for i in range(n)]
        self.__numItems = 0

    def __hash(self, x:int):
        return x % len(self.__table)

    def insert(self, x):
        slot = self.__hash(x)
        self.__table[slot].insert(0, x)
        self.__numItems += 1

    def search(self, x):
        slot = self.__hash(x)
        if len(self.__table[slot]) == 0:
            return None # null list
        else:
            i = self.__table[slot].index(x)
            if i < 0:
                return None # not exist
            else:
                return (self.__table[slot], i)

    def delete(self, x):
        if self.isEmpty():
            """ 에러 처리 """
        else:
            slot = self.__hash(x)
            self.__table[slot].remove(x)
            self.__numItems -= 1

    def isEmpty(self) -> bool:
        return self.__numItems == 0

    def clear(self):
        self.__table = [[] for i in range(len(self.__table))]
        self.__numItems = 0

    def printAll(self):
        for i in range(len(self.__table)):
            print("Slot ", i, " : ", end=" ")
            if len(self.__table[i]) == 0:
                print("Empty")
            else:
                for j in range(len(self.__table[i])):
                    print(self.__table[i][j], end=" ")
                print()

print("List Hash Table Demo!")
h = ListHashTable(11)
h.insert(10)
h.insert(21)
#h.delete(20)   # 아무 영향 없어야 함
h.insert(20)
h.insert(5)
h.insert(80)
h.insert(32)
h.printAll()

h.delete(20)
h.delete(80)

# Just for checking purpose
item = 21
(i, j) = h.search(item)
if i == None:
    print("Search Failed!")
else:
    print("Search:", item, "List:", i, ", item index", j)


from stack.listStack import *

def evaluate(p):
	s = ListStack()
	digitPreviously = False
	for i in range(len(p)):
		ch = p[i] #  i번 문자. 번호는 0번부터
		if ch.isdigit():	# ch가 숫자
			if digitPreviously:
				tmp = s.pop()
				tmp = 10 * tmp + (ord(ch) - ord('0'))
				s.push(tmp)
			else:
				s.push(ord(ch) - ord('0'))
				digitPreviously = True
		elif isOperator(ch): # ch가 연산자
			s.push(operation(s.pop(), s.pop(), ch))
			digitPreviously = False
		else:		# ch가 공백
			digitPreviously = False
	return s.pop()

def isOperator(ch) -> bool:  # 연산자인가?
	return (ch == '+' or ch == '-' or ch == '*' or ch == '/')

def operation(opr2:int, opr1:int, ch) -> int:  # 연산하기
		return {'+': opr1 + opr2, '-': opr1 - opr2, '*': opr1 * opr2, '/': opr1 // opr2}[ch]

def main():
	postfix = "700 3 47 + 6 * - 4 /"  # 테스트 샘플 입력(후위 표현식)
	print("Input string: ", postfix);
	answer = evaluate(postfix)
	print("Answer: ", answer)
	print(ord('0'), ord('9'))

if __name__ == "__main__":
	main()

# 코드 6-16


class HashOpenAddressed:
	def __init__(self, n:int):
		self.__table = [None for i in range(n)]
		self.__numItems = 0
		self.__DELETED = -54321

	def __hash(self, i:int, x):
		return (x + i) % len(self.__table)
 
	def insert(self, x):
		if self.__numItems == len(self.__table):
			""" 에러 처리 """
		else:
			for i in range(len(self.__table)):
				slot = self.__hash(i, x)
				if self.__table[slot] == None or self.__table[slot] == self.__DELETED:
					self.__table[slot] = x
					self.__numItems += 1
					break

	NOT_FOUND = -12345  # constant
	def search(self, x) -> int:
		for i in range(len(self.__table)):
			slot = self.__hash(i, x)
			if self.__table[slot] == None:
				return HashOpenAddressed.NOT_FOUND
			if self.__table[slot] == x:
				return slot  # Found at self.__table[slot]
		return self.__NOT_FOUND

	def delete(self, x):
		for i in range(len(self.__table)):
			slot = self.__hash(i, x)
			if self.__table[slot] == None:
				break
			if self.__table[slot] == x:
				self.__table[slot] = self.__DELETED;
				self.__numItems -= 1
				break

# 코드 12-3


from stack.listStack import *

def reverse(str):
	st = ListStack()
	for i in range(len(str)):
		st.push(str[i])
	out = ""
	while not st.isEmpty():
		out += st.pop()
	return out

def main():
	input = "Test Seq 12345"  # 테스트 입력 문자열 String
	answer = reverse(input)
	print("Input string: ", input)
	print("Reversed string: ", answer)

if __name__ == "__main__":
	main()

# 코드 6-15


from stack.listStack import *
from queue_.listQueue import *

def isPalindrome(A) -> bool:
	s = ListStack(); q = ListQueue()
	for i in range(len(A)):
		s.push(A[i]); q.enqueue(A[i])
	while (not q.isEmpty()) and s.pop() == q.dequeue():
		{}
	if q.isEmpty():
		return True
	else:
		return False

def main():
	print("Palindrome Check!")
	str = 'lioninoil' # 테스트 문자열
	t = isPalindrome(str)
	print(str, " is Palindrome?: ", t)

if __name__ == "__main__":
	main()

# 코드 7-15


def selectionSort(A):
	for last in range(len(A)-1, 0, -1):
		k = theLargest(A, last)	# A[0...last] 중 가장 큰 수 A[k] 찾기
		A[k], A[last] =  A[last], A[k]

def theLargest(A, last:int) -> int:	# A[0...last]에서 가장 큰 수의 인덱스를 리턴한다
	largest = 0
	for i in range(last+1):
		if A[i] > A[largest]:
			largest = i
	return largest

# 코드 9-1


from BST.binarySearchTree import *

bst1 = BinarySearchTree()
bst1.insert(10)
bst1.insert(20)
bst1.insert(5)
bst1.insert(80)
bst1.insert(90)
bst1.insert(7550)
bst1.insert(30)
bst1.insert(77)
bst1.insert(15)
bst1.insert(40)
bst1.delete(7550)
bst1.delete(10)

# 코드 10-2


def countingSort(A):
	k = max(A)	# 리스트 A[...]에서 최대값
	C = [0 for _ in range(k+1)]
	for j in range(len(A)):
		C[A[j]] += 1
	for i in range(1, k+1):
		C[i] = C[i] + C[i-1]
	B = [0 for _ in range(len(A))] 
	for j in range(len(A)-1, -1, -1):
		B[C[A[j]]-1] = A[j]
		C[A[j]] -= 1
	return B

# 코드 9-9


from list.circularLinkedList import CircularLinkedList

class LinkedQueue:
	def __init__(self):
		self.__queue = CircularLinkedList()

	def enqueue(self, x):
		self.__queue.append(x)

	def dequeue(self):
		return self.__queue.pop(0)	# .pop(0): 리스트의 첫 원소를 삭제한 후 원소 리턴

	def front(self):
		return self.__queue.get(0)	 # .get(0): 리스트의 첫 원소 리턴

	def isEmpty(self) -> bool:
		return self.__queue.isEmpty()
 
	def dequeueAll(self):
		self.__queue.clear()
	def printQueue(self):
		print("Queue from front:", end = ' ')
		for i in range(self.__queue.size()):
			print(self.__queue.get(i), end = ' ')
		print()

# 코드 7-13


import setuptools

setuptools.setup(
    include_package_data = True,
    name = "dsdau",
    version = "0.0.1.2",
    description = "DAU Data Structure Modules",
    author = "SangKyunJeon",
    author_email = "sangkyun.jeon@gmail.com",
    url = "https://github.com/jsk0910/DAU_DS",
    download_url = "https://github.com/jsk0910/DAU_DS/archive/refs/tags/v0.0.1.2.zip",
    install_requires = ['pytest'],
    py_modules = ['list.bidirectNode', 
                'list.circularDoublyLinkedList', 
                'list.circularDoublyLinkedListDemo', 
                'list.circularLinkedList',
                'list.circularLinkedListDemo',
                'list.linkedListBasic',
                'list.linkedListBasicDemo',
                'list.listNode',
                'stack.linkedStack',
                'stack.linkedStackDemo',
                'stack.listStack',
                'stack.listStackDemo',
                'stack.postfix',
                'stack.reverseString',
                'sort.bubbleSort',
                'sort.bubbleSortDemo',
                'sort.bucketSort',
                'sort.bucketSortDemo',
                'sort.countingSort',
                'sort.countingSortDemo',
                'sort.heapSort',
                'sort.heapSortDemo',
                'sort.insertionSort',
                'sort.insertionSortDemo',
                'sort.mergeSort',
                'sort.mergeSortDemo',
                'sort.quickSort',
                'sort.quickSortDemo',
                'sort.radixSort',
                'sort.radixSortDemo',
                'sort.selectionSort',
                'sort.selectionSortDemo',
                'sort.shellSort',
                'sort.shellSortDemo',
                'queue_.linkedQueue',
                'queue_.linkedQueueDemo',
                'queue_.listQueue',
                'queue_.listQueueDemo',
                'queue_.palindrome',
                'heap_.heap',
                'heap_.heapDemo',
                'hashTable.chainedHashTable',
                'hashTable.chainedHashTableDemo',
                'hashTable.hashOpenAddressed',
                'hashTable.hashOpenAddressedDemo',
                'hashTable.listHashTable',
                'hashTable.listHashTableDemo',
                'BST.AVLTree',
                'BST.AVLTreeDemo',
                'BST.binarySearchTree',
                'BST.binarySearchTreeDemo'
                ],
    package_dir = {"": "DS"},
    packages=setuptools.find_packages(where='DS'),
    long_description = "Dong-A Univ. Data Structure Lecture Modules",
    long_description_content_type = "text/markdown",
    classifiers= [
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)


class TreeNode:
	def __init__(self, newItem, left, right):
		self.item = newItem
		self.left = left
		self.right = right

class BinarySearchTree:
	def __init__(self):
		self.__root = None

	# [알고리즘 10-1] 구현: 검색
	def search(self, x) -> TreeNode:
		return self.__searchItem(self.__root, x)

	def __searchItem(self, tNode:TreeNode, x) -> TreeNode:
		if (tNode == None):
			return None
		elif (x == tNode.item):
			return tNode
		elif (x < tNode.item):
			return self.__searchItem(tNode.left, x)
		else:
			return self.__searchItem(tNode.right, x)

	# [알고리즘 10-3] 구현: 삽입
	def insert(self, newItem):
		self.__root = self.__insertItem(self.__root, newItem)

	def __insertItem(self, tNode:TreeNode, newItem) -> TreeNode:
		if (tNode == None):
			tNode = TreeNode(newItem, None, None)
		elif (newItem < tNode.item):  # branch left
			tNode.left = self.__insertItem(tNode.left, newItem)
		else:					     # branch right
			tNode.right = self.__insertItem(tNode.right, newItem)
		return tNode

	# [알고리즘 10-7] 구현: 삭제
	def delete(self, x):
		self.__root = self.__deleteItem(self.__root, x)
	
	def __deleteItem(self, tNode:TreeNode, x) -> TreeNode:
		if (tNode == None):
			return None			# Error! Item not found
		elif (x == tNode.item):	# item found!
				tNode = self.__deleteNode(tNode)
		elif (x < tNode.item):
				tNode.left = self.__deleteItem(tNode.left, x)
		else:
				tNode.right = self.__deleteItem(tNode.right, x)
		return tNode # tNode: parent에 매달리는 노드

	def __deleteNode(self, tNode:TreeNode) -> TreeNode:
		# 3가지 case
		#     1. tNode이 리프 노드
		#     2. tNode이 자식이 하나만 있음
		#     3. tNode이 자식이 둘 있음
		if tNode.left == None and tNode.right == None: # case 1(자식이 없음)
			return None
		elif tNode.left == None:  # case 2(오른자식뿐)
			return tNode.right
		elif tNode.right == None: # case 2(왼자식뿐)
			return tNode.left
		else: # case 3(두 자식이 다 있음)
			(rtnItem, rtnNode) = self.__deleteMinItem(tNode.right)
			tNode.item = rtnItem
			tNode.right = rtnNode
			return tNode  # tNode survived

	def __deleteMinItem(self, tNode:TreeNode) -> tuple:
		if tNode.left == None:
			# found min at tNode
			return (tNode.item, tNode.right)
		else: # branch left
			(rtnItem, rtnNode) = self.__deleteMinItem(tNode.left)
			tNode.left = rtnNode
			return (rtnItem, tNode)

	# 기타
	def isEmpty(self) -> bool:
		return self.__root == self.NIL

	def clear(self):
		self.__root = self.NIL

# 코드 10-1


from list.circularLinkedList import *
from list.listNode import *

class ChainedHashTable:
	def __init__(self, n):
		self.__table = [CircularLinkedList() for i in range(n)]
		self.__numItems = 0;

	def __hash(self, x:int): # 편의상 int 타입으로 제한
		return x % len(self.__table)

	# [알고리즘 12-1] 구현: 검색, 삽입, 삭제
	def insert(self, x:int):
		slot = self.__hash(x)
		self.__table[slot].insert(0, x)
		self.__numItems += 1

	def search(self, x:int) -> ListNode:
		slot = self.__hash(x)
		if self.__table[slot].isEmpty():
			return None # null list
		else:
			head = prev = self.__table[slot].getNode(-1)  # 더미 헤드
			curr = prev.next  # 0번 노드
			while curr != head:
				if curr.item == x:
					return curr
				else:
					prev = curr; curr = curr.next
			return None
	def delete(self, x:int):
		slot = self.__hash(x)
		success = self.__table[slot].remove(x)
		if success != None:
			self.__numItems -= 1
 	# 기타
	def isEmpty(self):
		return self.__numItems == 0

	def clear(self):
		for i in range(len(self.__table)):
			self.__table[i] = CircularLinkedList()
		self.__numItems = 0

# 코드 12-1


from stack.listStack import *

st1 = ListStack()
print(st1.top())	# No effect
st1.push(100)
st1.push(200)
print("Top is", st1.top())
st1.pop()
st1.push('Monday')
st1.printStack()
print('isEmpty?', st1.isEmpty())

# 코드 6-7


from list.bidirectNode import BidirectNode

class CircularDoublyLinkedList:
	def __init__(self):
		self.__head = BidirectNode("dummy", None, None)
		self.__head.prev = self.__head
		self.__head.next = self.__head
		self.__numItems = 0
 
	def insert(self, i:int, newItem) -> None:
		if (i >= 0 and i <= self.__numItems):
			prev = self.getNode(i - 1)
			newNode = BidirectNode(newItem, prev, prev.next)
			newNode.next.prev = newNode
			prev.next = newNode
			self.__numItems += 1
		else:
			print("index", i, ": out of bound in insert()") # 필요 시 에러 처리

	def append(self, newItem) -> None:
		prev = self.__head.prev
		newNode = BidirectNode(newItem, prev, self.__head)
		prev.next = newNode
		self.__head.prev = newNode
		self.__numItems += 1

	def pop(self, *args):
		# 가변 파라미터. 인자가 없거나 -1이면 마지막 원소로 처리하기 위함. 파이썬 리스트 규칙 만족
		if self.isEmpty():
			return None
		# 인덱스 i 결정
		if len(args) != 0: # pop(k)과 같이 인자가 있으면 i = k 할당
			i = args[0]
		if len(args) == 0 or i == -1:# pop()에 인자가 없거나 pop(-1)이면 i에 맨 끝 인덱스 할당
			i = self.__numItems - 1
		# i번 원소 삭제
		if (i >= 0 and i <= self.__numItems - 1):
			curr = self.getNode(i)
			retItem = curr.item
			curr.prev.next = curr.next
			curr.next.prev = curr.prev
			self.__numItems -= 1
			return retItem
		else:
			return None
 
	def remove(self, x):
		curr = self.__findNode(x)
		if curr != None:
			curr.prev.next = curr.next
			curr.next.prev = curr.prev
			self.__numItems -= 1
			return x
		else:
			return None

	def get(self, *args):
		# 가변 파라미터. 인자가 없거나 -1이면 마지막 원소로 처리하기 위함. 파이썬 리스트 규칙 만족
		if self.isEmpty(): return None
		# 인덱스 i 결정
		if len(args) != 0:   # pop(k)과 같이 인자가 있으면 i = k 할당
			i = args[0]
		if len(args) == 0 or i == -1:# pop()에 인자가 없거나 pop(-1)이면 i에 맨 끝 인덱스 할당
			i = self.__numItems - 1
		# i번 원소 리턴
		if (i >= 0 and i <= self.__numItems - 1):
			return self.getNode(i).item
		else:
			return None
 
	def index(self, x) -> int:
		cnt = 0
		for element in self:
			if element == x:
				return cnt
			cnt += 1
		return -12345

	def isEmpty(self) -> bool:
		return self.__numItems == 0
 
	def size(self) -> int:
		return self.__numItems

	def clear(self):
		self.__head = BidirectNode("dummy", None, None)
		self.__head.prev = self.__head
		self.__head.next = self.__head
		self.__numItems = 0
 
	def count(self, x) -> int:
		cnt = 0
		for element in self:
			if element == x:
					cnt += 1
		return cnt

	def extend(self, a): # a는 순회 가능한 모든 객체
		for element in a:
			self.append(element)

	def copy(self) -> 'CircularDoublyLinkedList':
		a = CircularDoublyLinkedList()
		for element in self:
			a.append(element)
		return a
 
	def reverse(self) -> None:
		prev = self.__head; curr = prev.next; next = curr.next
		self.__head.next = prev.prev; self.__head.prev = curr
		for i in range(self.__numItems):
			curr.next = prev; curr.prev = next
			prev = curr; curr = next; next = next.next
 
	def sort(self) -> None:
		a = []
		for element in self:
			a.append(element)
		a.sort()
		self.clear()
		for element in a:
			self.append(element)

	def __findNode(self, x) -> BidirectNode:
		curr = self.__head.next  # 0번 노드
		while curr != self.__head:
 
			if curr.item == x:
				return curr
			else:
				curr = curr.next
		return None

	def getNode(self, i:int) -> BidirectNode:
		curr = self.__head  # 더미 헤드, index: -1
		for index in range(i + 1):
			curr = curr.next
		return curr

	def printList(self) -> None:
		for element in self:
			print(element, end = ' ')
		print()
	def __iter__(self):  # generating iterator and return
		return CircularDoublyLinkedListIterator(self)
 
class CircularDoublyLinkedListIterator:
	def __init__(self, alist):
		self.__head = alist.getNode(-1)  		# 더미 헤드
		self.iterPosition = self.__head.next  	# 0번 노드
	def __next__(self):
		if self.iterPosition == self.__head: 	# 순환 끝
			raise StopIteration
		else: # 현재 원소를 리턴하면서 다음 원소로 이동
			item = self.iterPosition.item
			self.iterPosition = self.iterPosition.next
			return item

# 코드 5-25


from hashTable.chainedHashTable import *
from list.circularLinkedList import *
from list.listNode import *

h = ChainedHashTable(7)
h.insert(10)
h.insert(21)
h.delete(20) # 아무 영향 없어야 함
h.insert(20)
h.insert(5)
h.insert(12)
h.insert(19)
h.insert(80)
h.insert(32)
h.delete(20)
h.delete(44)

item = 21
slot = h.search(item)
if slot == None:
	print("Search Failed for", item)
else:
	print("Search for", item, "successful at slot", slot)

# 코드 12-2